unit ReservedWordsUnit;

interface

uses ComCtrls, Classes, SysUtils, Graphics, ReuUnit;

type

TReservedWords = object
private
  function GetWholeWord : boolean;
  procedure SetWholeWord(ww : boolean);
  function GetMatchCase : boolean;
  procedure SetMatchCase(mc : boolean);

protected
  FOptions : TSearchTypes;

public
  Items : TStringList;

  constructor Create;
  destructor Free;
  procedure SaveToFile(var FileVar : TextFile);
  procedure LoadFromFile(var FileVar : TextFile);

  property Options : TSearchTypes
    read FOptions;

  property WholeWord : boolean
    read GetWholeWord
    write SetWholeWord;

  property MatchCase : boolean
    read GetMatchCase
    write SetMatchCase;
end;

TLangReservedWords = object
public
  Extensions : TStringList;
  KeyWords : TReservedWords;
  Symbols :  TReservedWords;

  constructor Create;
  destructor Free;
  function HasExtension(const ext : TFileName) : boolean;
  procedure LoadFromFile(var FileVar : TextFile);
  procedure SaveToFile(var FileVar : TextFile);
end;

TLangReservedWordsPtr = ^TLangReservedWords;

TToken = (tkKeyWord, tkSymbol, tkPlainText, tkWhiteSpace);

TParseRecord = object
  Kind : TToken;
  Start, Length : integer;
end;
TParseRecordPtr = ^TParseRecord;

TParseList = class(TList)
private
  function GetItem(i : integer) : TParseRecordPtr;
  procedure SetItem(i : integer; parsePtr : TParseRecordPtr);
public
  property Items[i : integer] : TParseRecordPtr
    read GetItem
    write SetItem; default;

  function First : TParseRecordPtr;
  function Last : TParseRecordPtr;
end;

TParses = object
private
  MaxSymbolLength : integer;
  FLangWordsPtr : TLangReservedWordsPtr;

  procedure SetLangWordsPtr(lwp : TLangReservedWordsPtr);
  procedure OptimizeParse;

public
  Items : TParseList;

  property LangWordsPtr : TLangReservedWordsPtr
    read FLangWordsPtr
    write SetLangWordsPtr;

  procedure ParseLine(const line : string);
  constructor Create;
  destructor Free;
end;

TLangReservedWordsList = class(TList)
private
  FCurIndex : integer;
  procedure SetCurIndex(ci : integer);

public
  KeyWordsColor, SymbolsColor : TColor;
  Parses : TParses;

  property CurIndex : integer read FCurIndex write SetCurIndex;

  function GetItem(i : integer) : TLangReservedWordsPtr;
  property Items[i : integer] : TLangReservedWordsPtr
    read GetItem; default;

  function CurItem : TLangReservedWordsPtr;
  function GetItemIndex(const ext : TFileName) : integer;
  function UpdateCurItem(const fName : TFileName) : boolean;
  procedure LoadFromFile(var FileVar : TextFile);
  procedure SaveToFile(var FileVar : TextFile);
  constructor Create;
end;

function NewParseRecord(start, length : integer; kind : TToken) : TParseRecordPtr;

implementation

constructor TReservedWords.Create;
begin
  MatchCase := false;
  Items := TStringList.Create;
end;

destructor TReservedWords.Free;
begin
  Items.Free;
end;

procedure TReservedWords.SaveToFile(var FileVar : TextFile);
var i : integer;
begin
  WriteLn(FileVar, Ord(stMatchCase in Options));
  WriteLn(FileVar, Ord(stWholeWord in Options));

  WriteLn(FileVar, Items.Count);
  if Items.Count > 0 then
    for i := 0 to Items.Count - 1 do
      WriteLn(FileVar, Items[i]);
end;

procedure TReservedWords.LoadFromFile(var FileVar : TextFile);
var
  i : integer;
  inStr : string;
begin
  ReadLn(FileVar, i);
  MatchCase := IntToBool(i);
  ReadLn(FileVar, i);
  WholeWord := IntToBool(i);

  ReadLn(FileVar, i);
  Items.Clear;
  while (Items.Count < i) do begin
    ReadLn(FileVar, inStr);
    Items.Add(inStr);
  end;
end;

function TReservedWords.GetWholeWord : boolean;
begin
  result := stWholeWord in Options;
end;

procedure TReservedWords.SetWholeWord(ww : boolean);
begin
  if ww then
    FOptions := FOptions + [stWholeWord]
  else
    FOptions := FOptions - [stWholeWord];
end;

function TReservedWords.GetMatchCase : boolean;
begin
  result := stMatchCase in Options;
end;

procedure TReservedWords.SetMatchCase(mc : boolean);
begin
  if mc then
    FOptions := FOptions + [stMatchCase]
  else
    FOptions := FOptions - [stMatchCase];
end;

constructor TLangReservedWords.Create;
begin
  Extensions := TStringList.Create;
  KeyWords.Create;
  KeyWords.WholeWord := true;
  Symbols.Create;
  Symbols.WholeWord := false;
end;

destructor TLangReservedWords.Free;
begin
  Extensions.Free;
  KeyWords.Free;
  Symbols.Free;
end;

function TLangReservedWords.HasExtension(const ext : TFileName) : boolean;
begin
  result := Extensions.IndexOf(ext) <> -1;
end;

procedure TLangReservedWords.SaveToFile(var FileVar : TextFile);
var i : integer;
begin
  WriteLn(FileVar, Extensions.Count);
  if Extensions.Count > 0 then
    for i := 0 to Extensions.Count - 1 do
      WriteLn(FileVar, Extensions[i]);

  KeyWords.SaveToFile(FileVar);
  Symbols.SaveToFile(FileVar);
end;

procedure TLangReservedWords.LoadFromFile(var FileVar : TextFile);
var
  i : integer;
  inStr : string;
begin
  ReadLn(FileVar, i);
  while (Extensions.Count < i) do begin
    ReadLn(FileVar, inStr);
    Extensions.Add(inStr);
  end;

  KeyWords.LoadFromFile(FileVar);
  Symbols.LoadFromFile(FileVar);
end;

function TParseList.GetItem(i : integer) : TParseRecordPtr;
begin
  result := TParseRecordPtr(List[i]);
end;

procedure TParseList.SetItem(i : integer; parsePtr : TParseRecordPtr);
begin
  List[i] := parsePtr;
end;

function TParseList.First : TParseRecordPtr;
begin
  assert(Count > 0);
  result := Items[0];
end;

function TParseList.Last : TParseRecordPtr;
begin
  assert(Count > 0);
  result := Items[Count - 1];
end;

procedure TParses.SetLangWordsPtr(lwp : TLangReservedWordsPtr);
var i : integer;
begin
  FLangWordsPtr := lwp;
  MaxSymbolLength := 0;

  if LangWordsPtr <> nil then
    with LangWordsPtr^.Symbols do
      if Items.Count > 0 then
        for i := 0 to Items.Count - 1 do
          MaxSymbolLength := max(MaxSymbolLength, length(Items[i]));
end;

procedure TParses.OptimizeParse;
var i, j, k : integer;
begin
  i := Items.Count - 1;
  while (i > 0) do begin
    while (Items[i].Kind = tkWhiteSpace) and
          (i > 0) do
      dec(i);
    if i = 0 then exit;

    {this loop should find a j whose item has the same
     kind as i, and i and j are separated only by
     white spaces or items of that kind}
    j := i;
    k := i - 1;
    while ((Items[k].Kind = tkWhiteSpace) or
           (Items[i].Kind = Items[k].Kind)) and
          (k > 0) do begin
      if Items[i].Kind = Items[k].Kind then j := k;
      dec(k);
    end;
    assert(Items[i].Kind = Items[j].Kind);

    {if an appropriate j is found, then consider all
     items from j to i as one element of that kind}
    if j <> i then begin
      Items[j].Length := Items[i].Start + Items[i].Length - Items[j].Start;
      for k := j + 1 to i do
        Items.Delete(j + 1);

      i := Items.Count - 1;
    end else dec(i);
  end;
end;

procedure TParses.ParseLine(const line : string);
var pos, len : integer;

  function ParseName : boolean;
  begin
    result := true;
    if not IsAlpha(line[pos]) and
    not (line[pos] = '_') then exit;
    Items.Add(NewParseRecord(pos, 1, tkPlainText));
    inc(pos);

    while (pos <= len)
    and (isAlphaNum(line[pos]) or (line[pos] = '_')) do begin
      inc(pos);
      inc(Items.Last^.Length);
    end;

    if IndexOf(LangWordsPtr^.KeyWords.Items, copy(line, Items.Last^.Start, Items.Last^.Length), LangWordsPtr^.KeyWords.MatchCase) <> -1 then
      Items.Last^.Kind := tkKeyWord;

    result := pos <= len;
  end;

  function ParseSymbol : boolean;
  var i : integer;
  begin
    result := true;
    if MaxSymbolLength = 0 then exit;

    {min(len-pos+1... is used since there may not be
     MaxSymbolLength left of line}
    assert(min(len - pos + 1, MaxSymbolLength) >= 1);
    for i := min(len - pos + 1, MaxSymbolLength) downto 1 do
      if IndexOf(LangWordsPtr^.Symbols.Items, copy(line, pos, i), LangWordsPtr^.Symbols.MatchCase) <> -1 then begin
        Items.Add(NewParseRecord(pos, i, tkSymbol));
        pos := pos + i;
        result := pos <= len;
        break;
      end;
  end;

  function ParseWhiteSpace : boolean;
  begin
    result := true;
    if not IsWhiteSpace(line[pos]) then exit;
    Items.Add(NewParseRecord(pos, 1, tkWhiteSpace));
    inc(pos);

    while (pos <= len) and IsWhiteSpace(line[pos]) do begin
      inc(pos);
      inc(Items.Last^.Length);
    end;

    result := pos <= len;
  end;

var lastPos : integer;

begin
  Items.Clear;
  Items.Capacity := 50;
  if LangWordsPtr = nil then exit;
  len := length(line);
  pos := 1;
  if pos > len then exit;

  lastPos := pos;
  while ParseName and ParseWhiteSpace and ParseSymbol do begin
    //if a char does not parse, make it tkPlainText
    if pos = lastPos then begin
      Items.Add(NewParseRecord(pos, 1, tkPlainText));
      inc(pos);
      if pos > len then break;
    end;

    lastPos := pos;
  end;

  OptimizeParse;
end;

constructor TParses.Create;
begin
  Items := TParseList.Create;
end;

destructor TParses.Free;
begin
  Items.Free;
end;

procedure TLangReservedWordsList.SetCurIndex(ci : integer);
begin
  FCurIndex := ci;
  Parses.LangWordsPtr := CurItem;
end;

function TLangReservedWordsList.GetItem(i : integer) : TLangReservedWordsPtr;
begin
  if (i < 0) or (i >= Count) then
    result := nil
  else
    result := TLangReservedWordsPtr(Get(i));
end;

function TLangReservedWordsList.CurItem : TLangReservedWordsPtr;
begin
  result := Items[CurIndex];
end;

function TLangReservedWordsList.GetItemIndex(const ext : TFileName) : integer;
begin
  if Count > 0 then
    for result := 0 to Count - 1 do
      if Items[result].HasExtension(ext) then
        exit;

  result := -1;
end;

{sets CurIndex to value of GetItemIndex called with the
 extension of fName; returns true iff CurIndex changed}
function TLangReservedWordsList.UpdateCurItem(const fName : TFileName) : boolean;
var
  ext : TFileName;
  i : integer;
begin
  ext := ExtractFileExt(fName);
  //remove ext's initial .
  if pos('.', ext) = 1 then
    system.delete(ext, 1, 1);

  i := GetItemIndex(ext);
  result := CurIndex <> i;
  CurIndex := i;
end;

procedure TLangReservedWordsList.LoadFromFile(var FileVar : TextFile);
var
  i : integer;
  LangWordsPtr : ^TLangReservedWords;
begin
  ReadLn(FileVar, KeyWordsColor);
  ReadLn(FileVar, SymbolsColor);

  //ReadLn(FileVar, FCurIndex);
  //CurIndex := -1;

  ReadLn(FileVar, i);
  while (Count < i) do begin
    new(LangWordsPtr);
    LangWordsPtr^.Create;
    LangWordsPtr^.LoadFromFile(FileVar);
    Add(LangWordsPtr);
  end;
end;

procedure TLangReservedWordsList.SaveToFile(var FileVar : TextFile);
var i : integer;
begin
  WriteLn(FileVar, KeyWordsColor);
  WriteLn(FileVar, SymbolsColor);

  //WriteLn(FileVar, CurIndex);
  WriteLn(FileVar, Count);
  if Count > 0 then
    for i := 0 to Count - 1 do
      Items[i].SaveToFile(FileVar);
end;

constructor TLangReservedWordsList.Create;
begin
  inherited;
  Parses.Create;
end;

function NewParseRecord(start, length : integer; kind : TToken) : TParseRecordPtr;
begin
  new(result);
  result.Start := start;
  result.Length := length;
  result.Kind := kind;
end;

end.
